"""
(Unofficial) SAD to XSuite Converter

Output Writer: Reference Shifts
"""

################################################################################
# Import Packages
################################################################################
from ._000_helpers import *

################################################################################
# Lattice File
################################################################################
def create_refshift_lattice_file_information(line, line_table):

    ########################################
    # Get information
    ########################################
    unique_xyshift_names    = []
    unique_zetashift_names  = []
    unique_xrotation_names  = []
    unique_yrotation_names  = []
    unique_srotation_names  = []

    unique_xyshift_variable_names   = []
    unique_zetashift_variable_names = []
    unique_xrotation_variable_names = []
    unique_yrotation_variable_names = []
    unique_srotation_variable_names = []

    for xyshift in line_table.rows[line_table.element_type == 'XYShift'].name:
        parentname      = get_parentname(xyshift)
        variablename    = get_variablename(xyshift)
        if parentname not in unique_xyshift_names:
            unique_xyshift_names.append(parentname)
            unique_xyshift_variable_names.append(variablename)

    for zshift in line_table.rows[line_table.element_type == 'ZetaShift'].name:
        parentname      = get_parentname(zshift)
        variablename    = get_variablename(zshift)
        if parentname not in unique_zetashift_names:
            unique_zetashift_names.append(parentname)
            unique_zetashift_variable_names.append(variablename)

    for xrotation in line_table.rows[line_table.element_type == 'XRotation'].name:
        parentname      = get_parentname(xrotation)
        variablename    = get_variablename(xrotation)
        if parentname not in unique_xrotation_names:
            unique_xrotation_names.append(parentname)
            unique_xrotation_variable_names.append(variablename)

    for yrotation in line_table.rows[line_table.element_type == 'YRotation'].name:
        parentname      = get_parentname(yrotation)
        variablename    = get_variablename(yrotation)
        if parentname not in unique_yrotation_names:
            unique_yrotation_names.append(parentname)
            unique_yrotation_variable_names.append(variablename)

    for srotation in line_table.rows[line_table.element_type == 'SRotation'].name:
        parentname      = get_parentname(srotation)
        variablename    = get_variablename(srotation)
        if parentname not in unique_srotation_names:
            unique_srotation_names.append(parentname)
            unique_srotation_variable_names.append(variablename)

    ########################################
    # Ensure there are reference shifts in the line
    ########################################
    if len(unique_xyshift_names) == 0 and \
            len(unique_zetashift_names) == 0 and \
            len(unique_xrotation_names) == 0 and \
            len(unique_yrotation_names) == 0 and \
            len(unique_srotation_names) == 0:
        return ""

    ########################################
    # Create Output string
    ########################################
    output_string   = f"""
############################################################
# Reference Shifts
############################################################
"""

    ########################################
    # XYShifts
    ########################################
    if len(unique_xyshift_names) != 0:
        output_string += f"""
########################################
# XYShifts
########################################"""

        for xyshift_name, xyshift_variable_name in zip(unique_xyshift_names, unique_xyshift_variable_names):
            output_string += f"""
env.new(
    name        = '{xyshift_name}',
    parent      = xt.XYShift,
    dx          = 'dx_{xyshift_variable_name}',
    dy          = 'dy_{xyshift_variable_name}')"""

        output_string += "\n"

    ########################################
    # ZetaShifts
    ########################################
    if len(unique_zetashift_names) != 0:
        output_string += f"""
########################################
# ZetaShifts
########################################"""

        for zetashift_name, zetashift_variable_name in zip(unique_zetashift_names, unique_zetashift_variable_names):
            output_string += f"""
env.new(
    name        = '{zetashift_name}',
    parent      = xt.ZetaShift,
    dzeta       = 'dz_{zetashift_variable_name}')"""

        output_string += "\n"

    ########################################
    # YRotations
    ########################################
    if len(unique_yrotation_names) != 0:
        output_string += f"""
########################################
# YRotations (CHI1)
########################################"""

        for yrotation_name, yrotation_variable_name in zip(unique_yrotation_names, unique_yrotation_variable_names):
            output_string += f"""
env.new(
    name        = '{yrotation_name}',
    parent      = xt.YRotation,
    angle       = 'chi1_{yrotation_variable_name}')"""

        output_string += "\n"

    ########################################
    # XRotations
    ########################################
    if len(unique_xrotation_names) != 0:
        output_string += f"""
########################################
# XRotations (CHI2)
########################################"""

        for xrotation_name, xrotation_variable_name in zip(unique_xrotation_names, unique_xrotation_variable_names):
            output_string += f"""
env.new(
    name        = '{xrotation_name}',
    parent      = xt.XRotation,
    angle       = 'chi2_{xrotation_variable_name}')"""

        output_string += "\n"

    ########################################
    # SRotations
    ########################################
    if len(unique_srotation_names) != 0:
        output_string += f"""
########################################
# SRotations (CHI3)
########################################"""

        for srotation_name, srotation_variable_name in zip(unique_srotation_names, unique_srotation_variable_names):
            output_string += f"""
env.new(
    name        = '{srotation_name}',
    parent      = xt.SRotation,
    angle       = 'chi3_{srotation_variable_name}')"""

    ########################################
    # Return
    ########################################
    output_string += "\n"
    return output_string

################################################################################
# Optics File
################################################################################
def create_refshift_optics_file_information(line, line_table):

    ########################################
    # Get information
    ########################################
    unique_xyshift_names    = []
    unique_zetashift_names  = []
    unique_xrotation_names  = []
    unique_yrotation_names  = []
    unique_srotation_names  = []

    unique_xyshift_variable_names   = []
    unique_zetashift_variable_names = []
    unique_xrotation_variable_names = []
    unique_yrotation_variable_names = []
    unique_srotation_variable_names = []

    for xyshift in line_table.rows[line_table.element_type == 'XYShift'].name:
        parentname      = get_parentname(xyshift)
        variablename    = get_variablename(xyshift)
        if parentname not in unique_xyshift_names:
            unique_xyshift_names.append(parentname)
            unique_xyshift_variable_names.append(variablename)

    for zshift in line_table.rows[line_table.element_type == 'ZetaShift'].name:
        parentname      = get_parentname(zshift)
        variablename    = get_variablename(zshift)
        if parentname not in unique_zetashift_names:
            unique_zetashift_names.append(parentname)
            unique_zetashift_variable_names.append(variablename)

    for xrotation in line_table.rows[line_table.element_type == 'XRotation'].name:
        parentname      = get_parentname(xrotation)
        variablename    = get_variablename(xrotation)
        if parentname not in unique_xrotation_names:
            unique_xrotation_names.append(parentname)
            unique_xrotation_variable_names.append(variablename)

    for yrotation in line_table.rows[line_table.element_type == 'YRotation'].name:
        parentname      = get_parentname(yrotation)
        variablename    = get_variablename(yrotation)
        if parentname not in unique_yrotation_names:
            unique_yrotation_names.append(parentname)
            unique_yrotation_variable_names.append(variablename)

    for srotation in line_table.rows[line_table.element_type == 'SRotation'].name:
        parentname      = get_parentname(srotation)
        variablename    = get_variablename(srotation)
        if parentname not in unique_srotation_names:
            unique_srotation_names.append(parentname)
            unique_srotation_variable_names.append(variablename)

    ########################################
    # Ensure there are reference shifts in the line
    ########################################
    if len(unique_xyshift_names) == 0 and \
            len(unique_zetashift_names) == 0 and \
            len(unique_xrotation_names) == 0 and \
            len(unique_yrotation_names) == 0 and \
            len(unique_srotation_names) == 0:
        return ""

    ########################################
    # Sort the lists
    ########################################
    unique_xyshift_variable_names, unique_xyshift_names = map(
        list, zip(*sorted(zip(unique_xyshift_variable_names, unique_xyshift_names))))
    unique_zetashift_variable_names, unique_zetashift_names = map(
        list, zip(*sorted(zip(unique_zetashift_variable_names, unique_zetashift_names))))
    unique_xrotation_variable_names, unique_xrotation_names = map(
        list, zip(*sorted(zip(unique_xrotation_variable_names, unique_xrotation_names))))
    unique_yrotation_variable_names, unique_yrotation_names = map(
        list, zip(*sorted(zip(unique_yrotation_variable_names, unique_yrotation_names))))
    unique_srotation_variable_names, unique_srotation_names = map(
        list, zip(*sorted(zip(unique_srotation_variable_names, unique_srotation_names))))

    ########################################
    # Create Output string
    ########################################
    output_string = f"""
    ############################################################
    # Reference Shifts
    ############################################################
"""

    ########################################
    # XYShifts
    ########################################
    if len(unique_xyshift_names) != 0:
        output_string += f"""
    ########################################
    # XYShifts
    ########################################"""

        for xyshift_name, xyshift_variable_name in zip(unique_xyshift_names, unique_xyshift_variable_names):

            dx  = line[xyshift_name].dx
            dy  = line[xyshift_name].dy
            
            if dx != 0:
                output_string += f"""
    {f'dx_{xyshift_variable_name}'}{' ' * (OUTPUT_STRING_SEP - len(f'dx_{xyshift_variable_name}') + 4)}{'= '}{dx:.12f},"""
            if dy != 0:
                output_string += f"""
    {f'dy_{xyshift_variable_name}'}{' ' * (OUTPUT_STRING_SEP - len(f'dy_{xyshift_variable_name}') + 4)}{'= '}{dy:.12f},"""

        output_string += "\n"

    ########################################
    # ZetaShifts
    ########################################
    if len(unique_zetashift_names) != 0:
        output_string += f"""
    ########################################
    # ZetaShifts
    ########################################"""

        for zetashift_name, zetashift_variable_name in zip(unique_zetashift_names, unique_zetashift_variable_names):

            dz  = line[zetashift_name].dzeta
            
            if dz != 0:
                output_string += f"""
    {f'dz_{zetashift_variable_name}'}{' ' * (OUTPUT_STRING_SEP - len(f'dz_{zetashift_variable_name}') + 4)}{'= '}{dz:.12f},"""

        output_string += "\n"

    ########################################
    # YRotations
    ########################################
    if len(unique_yrotation_names) != 0:
        output_string += f"""
    ########################################
    # YRotations (CHI1)
    ########################################"""

        for yrotation_name, yrotation_variable_name in zip(unique_yrotation_names, unique_yrotation_variable_names):

            chi1    = line[yrotation_name].angle

            if chi1 != 0:
                output_string += f"""
    {f'chi1_{yrotation_variable_name}'}{' ' * (OUTPUT_STRING_SEP - len(f'chi1_{yrotation_variable_name}') + 4)}{'= '}{chi1:.12f},"""

        output_string += "\n"

    ########################################
    # XRotations
    ########################################
    if len(unique_xrotation_names) != 0:
        output_string += f"""
    ########################################
    # XRotations (CHI2)
    ########################################"""

        for xrotation_name, xrotation_variable_name in zip(unique_xrotation_names, unique_xrotation_variable_names):

            chi2    = line[xrotation_name].angle

            if chi2 != 0:
                output_string += f"""
    {f'chi2_{xrotation_variable_name}'}{' ' * (OUTPUT_STRING_SEP - len(f'chi2_{xrotation_variable_name}') + 4)}{'= '}{chi2:.12f},"""

        output_string += "\n"

    ########################################
    # XRotations
    ########################################
    if len(unique_srotation_names) != 0:
        output_string += f"""
    ########################################
    # SRotations (CHI3)
    ########################################"""

        for srotation_name, srotation_variable_name in zip(unique_srotation_names, unique_srotation_variable_names):

            chi3    = line[srotation_name].angle

            if chi3 != 0:
                output_string += f"""
    {f'chi3_{srotation_variable_name}'}{' ' * (OUTPUT_STRING_SEP - len(f'chi3_{srotation_variable_name}') + 4)}{'= '}{chi3:.12f},"""

    ########################################
    # Return
    ########################################
    output_string += "\n"
    return output_string



"""
(Unofficial) SAD to XSuite Converter

Solenoid Converter
"""

################################################################################
# Required Packages
################################################################################
import numpy as np
import xtrack as xt

from tqdm import tqdm

from .._globals import print_section_heading, MAX_KNL_ORDER, N_MULTIPOLE_KICKS_SOL

################################################################################
# Conversion Function
################################################################################
def convert_solenoids(
        parsed_lattice_data:    dict,
        environment:            xt.Environment,
        verbose:                bool                = False) -> None:

    ########################################
    # Get the required data
    ########################################
    parsed_elements = parsed_lattice_data['elements']

    ########################################
    # Check if there are any solenoids
    ########################################
    if 'sol' not in parsed_elements:
        if verbose:
            print_section_heading("No solenoids in line", mode = 'subsection')
        return
    else:
        solenoids   = parsed_elements['sol']

    ########################################
    # Get bound and geo solenoids
    ########################################
    bound_solenoids = []
    geo_solenoids   = []
    for ele_name, ele_vars in solenoids.items():

        if 'bound' in ele_vars:
            bound_solenoids.append(ele_name)
        if 'geo' in ele_vars:
            geo_solenoids.append(ele_name)

    ############################################################################
    # Iterate through lines
    ############################################################################
    for line_name in environment.lines:

        # The line may be a compound solenoid element
        # e.g. dx, chi1, sol
        if line_name in bound_solenoids:
            continue
        if line_name.endswith("_reversed") and line_name[:-9] in bound_solenoids:
            continue

        line    = environment.lines[line_name]

        bound_sols_in_line      = []
        bound_solenoid_indicies = []
        for idx, element in enumerate(line.element_names):

            # Element must actually be a solenoid
            if not isinstance(environment.element_dict[element], xt.Solenoid):  # type: ignore
                continue

            # Case where the element is a bound solenoid with the right name
            if element in bound_solenoids:
                bound_sols_in_line.append(element)
                bound_solenoid_indicies.append(idx)
            # Case where the bound solenoid has the _bound suffix
            elif element.endswith("_bound") and element[:-6] in bound_solenoids:
                bound_sols_in_line.append(element)
                bound_solenoid_indicies.append(idx)
            # Same cases, but for reversed elements
            # Case where the element is a bound solenoid with the right name
            elif element[1:] in bound_solenoids:
                bound_sols_in_line.append(element)
                bound_solenoid_indicies.append(idx)
            # Case where the bound solenoid has the _bound suffix
            elif element.endswith("_bound") and element[1:-6] in bound_solenoids:
                bound_sols_in_line.append(element)
                bound_solenoid_indicies.append(idx)

        # If no bound solenoids are found in the line, skip to the next line
        if len(bound_sols_in_line) == 0:
            continue

        ########################################
        # Ensure an even number of boundary solenoids
        ########################################
        bound_solenoid_pairs            = []
        bound_solenoid_pair_indicies    = []
        for i in range(0, len(bound_sols_in_line), 2):
            if i + 1 < len(bound_sols_in_line):
                bound_solenoid_pairs.append((bound_sols_in_line[i], bound_sols_in_line[i + 1]))
                bound_solenoid_pair_indicies.append((bound_solenoid_indicies[i], bound_solenoid_indicies[i + 1]))
            else:
                raise ValueError("Unmatched solenoid found in the line.")

        ########################################
        # Get the elements between bound solenoids
        ########################################
        reversed_solenoid   = False
        previous_solenoid   = None
        ahead_solenoid      = None
        ks_previous         = 0
        ks_ahead            = 0
        segment_length      = 0
        ele_length          = 0
        s_start_ele         = 0
        s_end_ele           = 0


        INTERPOLATE         = False
        
        ########################################
        # Get the elements between bound solenoids
        ########################################
        for (sol_start, sol_end), (start_idx, end_idx) in zip(bound_solenoid_pairs, bound_solenoid_pair_indicies):

            if start_idx > end_idx:
                raise ValueError(f"Start solenoid {sol_start} is after end solenoid {sol_end} in line {line_name}.")

            ########################################
            # Loop through the elements between
            ########################################
            # Here, iterate through all the elements in the line
            # Enumerate also for the index
            # If the element index < start_index, it is before the solenoid
            # If the element index > end_index, it is after the solenoid
            # If the element index is between start_index and end_index, it is between the solenoids
            # Makes it easier to do the replacement
            for idx, element in enumerate(line.element_names):

                if idx < start_idx or idx >= end_idx:
                    # If the element is before or after the solenoid, skip it
                    continue

                # Do the swaps if the element is reversed
                if reversed_solenoid:
                    solenoid_suffix = ahead_solenoid
                    ks              = ks_ahead
                else:
                    solenoid_suffix = previous_solenoid
                    ks              = ks_previous

                # !!!!! INTERPOLATION MODE
                # # Calculate ks at the start and end of the element by interpolating
                if INTERPOLATE:

                    # Adjust the s_start based on the previous element
                    s_start_ele = s_end_ele
                    # Get the length of the current element
                    try:
                        ele_length  = line[element].length
                    except AttributeError:
                        # Some elements (like markers) have no length
                        ele_length  = 0
                    s_end_ele   = s_start_ele + ele_length
                    
                    if segment_length != 0:
                        ks_ele_start   = ((ks_ahead - ks_previous) / segment_length) * s_start_ele + ks_previous
                        ks_ele_end     = ((ks_ahead - ks_previous) / segment_length) * s_end_ele + ks_previous
                        # Mean at the centre of the element
                        ks              = (ks_ele_start + ks_ele_end) / 2
                    else:
                        ks_ele_start   = ks_previous
                        ks_ele_end     = ks_ahead
                        ks             = (ks_ele_start + ks_ele_end) / 2

                # If the element is a solenoid, update the current solenoidal field
                if isinstance(environment.element_dict[element], xt.Solenoid):  # type: ignore

                    # Start calculating new segment length
                    segment_length = ele_length
                    
                    # If the solenoid is reversed, we need to swap to that case
                    if element.startswith("-"):
                        reversed_solenoid    = True
                    else:
                        reversed_solenoid    = False

                    if element.endswith("_bound"):
                        # If the element is a bound solenoid clip this bit
                        previous_solenoid    = element[:-6]
                    else:
                        previous_solenoid    = element
                    ks_previous         = line[element].ks

                    # Get the information about the ahead solenoid
                    for ahead_element in line.element_names[idx + 1:]:
                        # If the element is a solenoid, update the solenoidal field
                        if isinstance(environment.element_dict[ahead_element], xt.Solenoid):    # type: ignore
                            if ahead_element.endswith("_bound"):
                                # If the element is a bound solenoid clip this bit
                                ahead_solenoid      = ahead_element[:-6]
                            else:
                                ahead_solenoid      = ahead_element
                            ks_ahead            = line[ahead_element].ks
                            break
                        else:
                            try:
                                segment_length += line[ahead_element].length
                            except AttributeError:
                                # Some elements (like markers) have no length
                                pass

                    # Reset the segment lengths as new segment
                    s_start_ele = 0
                    s_end_ele   = ele_length

                    continue

                # Drift conversion
                elif isinstance(environment.element_dict[element], xt.Drift):   # type: ignore

                    length              = line[element].length
                    new_element_name    = f"{element}_{solenoid_suffix}"

                    if INTERPOLATE:
                        rand_int            = np.random.randint(1, 1000)
                        new_element_name    = f"{element}_{solenoid_suffix}_{rand_int}"

                    if new_element_name not in environment.element_dict:        # type: ignore
                        environment.new(
                            name    = new_element_name,
                            parent  = xt.Solenoid,
                            length  = length,
                            ks      = ks)
                    line.element_names[idx] = new_element_name

                    if verbose:
                        print(f"Converted drift {element} to solenoid {new_element_name} with ks = {ks}")
                    continue

                # Bend conversion
                elif isinstance(environment.element_dict[element], xt.Bend):        # type: ignore

                    assert line[element].h == 0, "Bend with non-zero angle found between solenoids."

                    length      = line[element].length
                    k0          = line[element].k0
                    k1          = line[element].k1
                    shift_x     = line[element].shift_x
                    shift_y     = line[element].shift_y
                    rotation    = line[element].rot_s_rad
                    knl         = [f'{k0} * {length}', f'{k1} * {length}']

                    new_element_name    = f"{element}_{solenoid_suffix}"

                    if new_element_name not in environment.element_dict:        # type: ignore
                        environment.new(
                            name				= new_element_name,
                            parent				= xt.Solenoid,
                            length				= length,
                            ks					= ks,
                            knl					= knl,
                            num_multipole_kicks	= N_MULTIPOLE_KICKS_SOL,
                            order				= MAX_KNL_ORDER,
                            mult_shift_x		= shift_x,
                            mult_shift_y		= shift_y,
                            rot_s_rad           = rotation)
                    line.element_names[idx] = new_element_name

                    if verbose:
                        print(f"Converted Bend {element} to solenoid {new_element_name} with ks = {ks}")
                    continue
                    
                # Quadrupole conversion
                elif isinstance(environment.element_dict[element], xt.Quadrupole):   # type: ignore

                    length      = line[element].length
                    k1          = line[element].k1
                    k1s         = line[element].k1s
                    shift_x     = line[element].shift_x
                    shift_y     = line[element].shift_y
                    rotation    = line[element].rot_s_rad
                    knl         = [0, f'{k1} * {length}']
                    ksl         = [0, f'{k1s} * {length}']

                    new_element_name    = f"{element}_{solenoid_suffix}"

                    if new_element_name not in environment.element_dict:        # type: ignore
                        environment.new(
                            name				= new_element_name,
                            parent				= xt.Solenoid,
                            length				= length,
                            ks					= ks,
                            knl					= knl,
                            ksl					= ksl,
                            num_multipole_kicks	= N_MULTIPOLE_KICKS_SOL,
                            order				= MAX_KNL_ORDER,
                            mult_shift_x		= shift_x,
                            mult_shift_y		= shift_y,
                            rot_s_rad           = rotation)
                    line.element_names[idx] = new_element_name

                    if verbose:
                        print(f"Converted Quadrupole {element} to solenoid {new_element_name} with ks = {ks}")
                    continue
                    
                # Sextupole conversion
                elif isinstance(environment.element_dict[element], xt.Sextupole):   # type: ignore

                    length      = line[element].length
                    k2          = line[element].k2
                    k2s         = line[element].k2s
                    shift_x     = line[element].shift_x
                    shift_y     = line[element].shift_y
                    rotation    = line[element].rot_s_rad
                    knl         = [0, 0, f'{k2} * {length}']
                    ksl         = [0, 0, f'{k2s} * {length}']

                    new_element_name    = f"{element}_{solenoid_suffix}"

                    if new_element_name not in environment.element_dict:     # type: ignore
                        environment.new(
                            name				= new_element_name,
                            parent				= xt.Solenoid,
                            length				= length,
                            ks					= ks,
                            knl					= knl,
                            ksl					= ksl,
                            num_multipole_kicks	= N_MULTIPOLE_KICKS_SOL,
                            order				= MAX_KNL_ORDER,
                            mult_shift_x		= shift_x,
                            mult_shift_y		= shift_y,
                            rot_s_rad           = rotation)
                    line.element_names[idx] = new_element_name

                    if verbose:
                        print(f"Converted Sextupole {element} to solenoid {new_element_name} with ks = {ks}")
                    continue

                # Octupole conversion
                elif isinstance(environment.element_dict[element], xt.Octupole):    # type: ignore

                    length      = line[element].length
                    k3          = line[element].k3
                    k3s         = line[element].k3s
                    shift_x     = line[element].shift_x
                    shift_y     = line[element].shift_y
                    rotation    = line[element].rot_s_rad
                    knl         = [0, 0, 0, f'{k3} * {length}']
                    ksl         = [0, 0, 0, f'{k3s} * {length}']

                    new_element_name    = f"{element}_{solenoid_suffix}"

                    if new_element_name not in environment.element_dict:    # type: ignore
                        environment.new(
                            name				= new_element_name,
                            parent				= xt.Solenoid,
                            length				= length,
                            ks					= ks,
                            knl					= knl,
                            ksl					= ksl,
                            num_multipole_kicks	= N_MULTIPOLE_KICKS_SOL,
                            order				= MAX_KNL_ORDER,
                            mult_shift_x		= shift_x,
                            mult_shift_y		= shift_y,
                            rot_s_rad           = rotation)
                    line.element_names[idx] = new_element_name

                    if verbose:
                        print(f"Converted Octupole {element} to solenoid {new_element_name} with ks = {ks}")
                    continue

                # Multipole conversion
                elif isinstance(environment.element_dict[element], xt.Multipole):   # type: ignore

                    length      = line[element].length
                    knl         = line[element].knl
                    ksl         = line[element].ksl
                    shift_x     = line[element].shift_x
                    shift_y     = line[element].shift_y
                    rotation    = line[element].rot_s_rad

                    environment.element_dict.pop(element)                       # type: ignore
                    environment.new(
                        name				= element,
                        parent				= xt.Solenoid,
                        length				= length,
                        ks					= ks,
                        knl					= knl,
                        ksl					= ksl,
                        num_multipole_kicks	= N_MULTIPOLE_KICKS_SOL,
                        order				= MAX_KNL_ORDER,
                        mult_shift_x		= shift_x,
                        mult_shift_y		= shift_y,
                        rot_s_rad           = rotation)

                    if verbose:
                        print(f"Converted Multipole {element} to solenoid with ks = {ks}")
                    continue

                # Generic Magnet conversion (currently unused)
                elif isinstance(environment.element_dict[element], xt.Magnet):  # type: ignore

                    length      = line[element].length
                    knl         = line[element].knl
                    ksl         = line[element].ksl
                    shift_x     = line[element].shift_x
                    shift_y     = line[element].shift_y
                    rotation    = line[element].rot_s_rad

                    environment.element_dict.pop(element)                       # type: ignore
                    environment.new(
                        name				= element,
                        parent				= xt.Solenoid,
                        length				= length,
                        ks					= ks,
                        knl					= knl,
                        ksl					= ksl,
                        num_multipole_kicks	= N_MULTIPOLE_KICKS_SOL,
                        order               = MAX_KNL_ORDER,
                        mult_shift_x		= shift_x,
                        mult_shift_y		= shift_y,
                        rot_s_rad           = rotation)

                    if verbose:
                        print(f"Converted Multipole {element} to solenoid with ks = {ks}")
                    continue
                
                elif isinstance(
                    environment.element_dict[element],      # type: ignore
                    (
                        xt.XYShift,
                        xt.ZetaShift,
                        xt.XRotation,
                        xt.YRotation,
                        xt.SRotation,
                        xt.Marker,
                        xt.LimitEllipse)):  
                    # Known elements that don't need conversion
                    continue
                elif verbose:
                    print(f"Element {element} in line {line_name} has not been converted")

###############################################################################
# Reference shift corrections
###############################################################################
def solenoid_reference_shift_corrections(
        line:                   xt.Line,
        parsed_lattice_data:    dict,
        environment:            xt.Environment,
        reverse_line:           bool                = False,
        verbose:                bool                = False) -> None:

    ########################################
    # Get the required data
    ########################################
    parsed_elements = parsed_lattice_data['elements']

    ########################################
    # Check if there are any solenoids
    ########################################
    if 'sol' not in parsed_elements:
        if verbose:
            print_section_heading("No solenoids in line", mode = 'subsection')
        return
    else:
        solenoids   = parsed_elements['sol']

    ########################################
    # Get bound and geo solenoids
    ########################################
    bound_solenoids = []
    geo_solenoids   = []
    for ele_name, ele_vars in solenoids.items():

        if 'bound' in ele_vars:
            bound_solenoids.append(ele_name)
        if 'geo' in ele_vars:
            geo_solenoids.append(ele_name)

    ########################################
    # Get bound solenoids in the line
    ########################################
    bound_sols_in_line  = []
    geo_sols_in_line    = []
    for element in line.element_names:
        
        # Element must actually be a solenoid
        if not isinstance(environment.element_dict[element], xt.Solenoid):  # type: ignore
            continue

        # Case where the element is a bound solenoid with the right name
        if element in bound_solenoids:
            bound_sols_in_line.append(element)
        # Case where the bound solenoid has the _bound suffix
        elif element.endswith("_bound") and element[:-6] in bound_solenoids:
            bound_sols_in_line.append(element)
        # Same cases, but for reversed elements
        # Case where the element is a bound solenoid with the right name
        elif element[1:] in bound_solenoids:
            bound_sols_in_line.append(element)
        # Case where the bound solenoid has the _bound suffix
        elif element.endswith("_bound") and element[1:-6] in bound_solenoids:
            bound_sols_in_line.append(element)

        # Case where the element is a geo solenoid with the right name
        if element in geo_solenoids:
            geo_sols_in_line.append(element)
        # Case where the geo solenoid has the _bound suffix
        elif element.endswith("_bound") and element[:-6] in geo_solenoids:
            geo_sols_in_line.append(element)
        # Same cases, but for reversed elements
        # Case where the element is a bound solenoid with the right name
        elif element[1:] in geo_solenoids:
            geo_sols_in_line.append(element)
        # Case where the geo solenoid has the _bound suffix
        elif element.endswith("_bound") and element[1:-6] in geo_solenoids:
            geo_sols_in_line.append(element)

    # If no bound solenoids are found in the line, return
    if len(bound_sols_in_line) == 0:
        return

    ########################################
    # Ensure an even number of boundary solenoids
    ########################################
    bound_solenoid_pairs    = []
    for i in range(0, len(bound_sols_in_line), 2):
        if i + 1 < len(bound_sols_in_line):
            bound_solenoid_pairs.append(
                (bound_sols_in_line[i], bound_sols_in_line[i + 1]))
        else:
            raise ValueError("Unmatched solenoid found in the line.")

    ############################################################################
    # Get the inbound and outbound boundary solenoids
    ############################################################################
    inbound_solenoids   = [pair[0] for pair in bound_solenoid_pairs]
    outbound_solenoids  = [pair[1] for pair in bound_solenoid_pairs]
    # For each of the pairs, one of them is a geomatric solenoid
    geometric_solenoids     = []
    non_geometric_solenoids = []
    for pair in bound_solenoid_pairs:
        if pair[0] in geo_sols_in_line:
            geometric_solenoids.append(pair[0])
            non_geometric_solenoids.append(pair[1])
        elif pair[1] in geo_sols_in_line:
            geometric_solenoids.append(pair[1])
            non_geometric_solenoids.append(pair[0])
        else:
            raise ValueError(f"Neither solenoid in pair {pair} is a geometric solenoid.")

    inbound_solenoids       = list(set(inbound_solenoids))
    outbound_solenoids      = list(set(outbound_solenoids))
    geometric_solenoids     = list(set(geometric_solenoids))
    non_geometric_solenoids = list(set(non_geometric_solenoids))

    # We only care about the compound solenoids (the ones with reference frame transforms)
    # These solenoids must have the _bound suffix
    inbound_solenoids       = [sol for sol in inbound_solenoids if sol.endswith('_bound')]
    outbound_solenoids      = [sol for sol in outbound_solenoids if sol.endswith('_bound')]
    geometric_solenoids     = [sol for sol in geometric_solenoids if sol.endswith('_bound')]
    non_geometric_solenoids = [sol for sol in non_geometric_solenoids if sol.endswith('_bound')]

    # Here we need to remove the _sol suffix for the reference shift correction
    inbound_solenoids       = [sol[:-6] for sol in inbound_solenoids]
    outbound_solenoids      = [sol[:-6] for sol in outbound_solenoids]
    geometric_solenoids     = [sol[:-6] for sol in geometric_solenoids]
    non_geometric_solenoids = [sol[:-6] for sol in non_geometric_solenoids]

    inbound_solenoids       = sorted(inbound_solenoids)
    outbound_solenoids      = sorted(outbound_solenoids)
    geometric_solenoids     = sorted(geometric_solenoids)
    non_geometric_solenoids = sorted(non_geometric_solenoids)

    if verbose:
        print_section_heading("Reference Shift Solenoids:", mode = 'subsection')
        print(f"Inbound solenoids with ref transforms: {inbound_solenoids}")
        print(f"Outbound solenoids with ref transforms: {outbound_solenoids}")
        print(f"Geometric solenoids with ref transforms: {geometric_solenoids}")
        print(f"Non-geometric solenoids with ref transforms: {non_geometric_solenoids}")

    ############################################################################
    # DXY and CHI for all cases (inbound, geo, reverse_self, reverse_other)
    ############################################################################
    inbound_geo_forward_forward_solenoids       = []
    inbound_geo_forward_reverse_solenoids       = []
    inbound_geo_reverse_forward_solenoids       = []
    inbound_geo_reverse_reverse_solenoids       = []
    inbound_nongeo_forward_forward_solenoids    = []
    inbound_nongeo_forward_reverse_solenoids    = []
    inbound_nongeo_reverse_forward_solenoids    = []
    inbound_nongeo_reverse_reverse_solenoids    = []
    outbound_geo_forward_forward_solenoids      = []
    outbound_geo_forward_reverse_solenoids      = []
    outbound_geo_reverse_forward_solenoids      = []
    outbound_geo_reverse_reverse_solenoids      = []
    outbound_nongeo_forward_forward_solenoids   = []
    outbound_nongeo_forward_reverse_solenoids   = []
    outbound_nongeo_reverse_forward_solenoids   = []
    outbound_nongeo_reverse_reverse_solenoids   = []
    
    for inbound_solenoid, outbound_solenoid in bound_solenoid_pairs:
        
        # Reversal information
        inbound_reversed    = False
        outbound_reversed   = False
        if inbound_solenoid.startswith('-'):
            inbound_reversed    = True
        if outbound_solenoid.startswith('-'):
            outbound_reversed   = True

        # Inbound solnoids

        # Only care about the ones with reference shift transformations (should always be true)
        if inbound_solenoid.endswith('_bound'):
            inbound_solenoid    = inbound_solenoid[:-6]
            if inbound_solenoid in geometric_solenoids:

                if (not inbound_reversed) and (not outbound_reversed):
                    inbound_geo_forward_forward_solenoids.append(inbound_solenoid)
                elif (not inbound_reversed) and outbound_reversed:
                    inbound_geo_forward_reverse_solenoids.append(inbound_solenoid)
                elif inbound_reversed and (not outbound_reversed):
                    inbound_geo_reverse_forward_solenoids.append(inbound_solenoid)
                elif inbound_reversed and outbound_reversed:
                    inbound_geo_reverse_reverse_solenoids.append(inbound_solenoid)
            else:

                if (not inbound_reversed) and (not outbound_reversed):
                    inbound_nongeo_forward_forward_solenoids.append(inbound_solenoid)
                elif (not inbound_reversed) and outbound_reversed:
                    inbound_nongeo_forward_reverse_solenoids.append(inbound_solenoid)
                elif inbound_reversed and (not outbound_reversed):
                    inbound_nongeo_reverse_forward_solenoids.append(inbound_solenoid)
                elif inbound_reversed and outbound_reversed:
                    inbound_nongeo_reverse_reverse_solenoids.append(inbound_solenoid)

        # Only care about the boundary solenoids
        if outbound_solenoid.endswith('_bound'):
            outbound_solenoid    = outbound_solenoid[:-6]

            if outbound_solenoid in geometric_solenoids:
                
                if (not inbound_reversed) and (not outbound_reversed):
                    outbound_geo_forward_forward_solenoids.append(outbound_solenoid)
                elif (not inbound_reversed) and outbound_reversed:
                    outbound_geo_forward_reverse_solenoids.append(outbound_solenoid)
                elif inbound_reversed and (not outbound_reversed):
                    outbound_geo_reverse_forward_solenoids.append(outbound_solenoid)
                elif inbound_reversed and outbound_reversed:
                    outbound_geo_reverse_reverse_solenoids.append(outbound_solenoid)
            else:
                
                if (not inbound_reversed) and (not outbound_reversed):
                    outbound_nongeo_forward_forward_solenoids.append(outbound_solenoid)
                elif (not inbound_reversed) and outbound_reversed:
                    outbound_nongeo_forward_reverse_solenoids.append(outbound_solenoid)
                elif inbound_reversed and (not outbound_reversed):
                    outbound_nongeo_reverse_forward_solenoids.append(outbound_solenoid)
                elif inbound_reversed and outbound_reversed:
                    outbound_nongeo_reverse_reverse_solenoids.append(outbound_solenoid)

    inbound_geo_forward_forward_solenoids       = list(set(inbound_geo_forward_forward_solenoids))
    inbound_geo_forward_reverse_solenoids       = list(set(inbound_geo_forward_reverse_solenoids))
    inbound_geo_reverse_forward_solenoids       = list(set(inbound_geo_reverse_forward_solenoids))
    inbound_geo_reverse_reverse_solenoids       = list(set(inbound_geo_reverse_reverse_solenoids))
    inbound_nongeo_forward_forward_solenoids    = list(set(inbound_nongeo_forward_forward_solenoids))
    inbound_nongeo_forward_reverse_solenoids    = list(set(inbound_nongeo_forward_reverse_solenoids))
    inbound_nongeo_reverse_forward_solenoids    = list(set(inbound_nongeo_reverse_forward_solenoids))
    inbound_nongeo_reverse_reverse_solenoids    = list(set(inbound_nongeo_reverse_reverse_solenoids))
    outbound_geo_forward_forward_solenoids      = list(set(outbound_geo_forward_forward_solenoids))
    outbound_geo_forward_reverse_solenoids      = list(set(outbound_geo_forward_reverse_solenoids))
    outbound_geo_reverse_forward_solenoids      = list(set(outbound_geo_reverse_forward_solenoids))
    outbound_geo_reverse_reverse_solenoids      = list(set(outbound_geo_reverse_reverse_solenoids))
    outbound_nongeo_forward_forward_solenoids   = list(set(outbound_nongeo_forward_forward_solenoids))
    outbound_nongeo_forward_reverse_solenoids   = list(set(outbound_nongeo_forward_reverse_solenoids))
    outbound_nongeo_reverse_forward_solenoids   = list(set(outbound_nongeo_reverse_forward_solenoids))
    outbound_nongeo_reverse_reverse_solenoids   = list(set(outbound_nongeo_reverse_reverse_solenoids))

    if verbose:
        print(f"inbound_geo_forward_forward_solenoids     = {inbound_geo_forward_forward_solenoids}")
        print(f"inbound_geo_forward_reverse_solenoids     = {inbound_geo_forward_reverse_solenoids}")
        print(f"inbound_geo_reverse_forward_solenoids     = {inbound_geo_reverse_forward_solenoids}")
        print(f"inbound_geo_reverse_reverse_solenoids     = {inbound_geo_reverse_reverse_solenoids}")
        print(f"inbound_nongeo_forward_forward_solenoids  = {inbound_nongeo_forward_forward_solenoids}")
        print(f"inbound_nongeo_forward_reverse_solenoids  = {inbound_nongeo_forward_reverse_solenoids}")
        print(f"inbound_nongeo_reverse_forward_solenoids  = {inbound_nongeo_reverse_forward_solenoids}")
        print(f"inbound_nongeo_reverse_reverse_solenoids  = {inbound_nongeo_reverse_reverse_solenoids}")
        print(f"outbound_geo_forward_forward_solenoids    = {outbound_geo_forward_forward_solenoids}")
        print(f"outbound_geo_forward_reverse_solenoids    = {outbound_geo_forward_reverse_solenoids}")
        print(f"outbound_geo_reverse_forward_solenoids    = {outbound_geo_reverse_forward_solenoids}")
        print(f"outbound_geo_reverse_reverse_solenoids    = {outbound_geo_reverse_reverse_solenoids}")
        print(f"outbound_nongeo_forward_forward_solenoids = {outbound_nongeo_forward_forward_solenoids}")
        print(f"outbound_nongeo_forward_reverse_solenoids = {outbound_nongeo_forward_reverse_solenoids}")
        print(f"outbound_nongeo_reverse_forward_solenoids = {outbound_nongeo_reverse_forward_solenoids}")
        print(f"outbound_nongeo_reverse_reverse_solenoids = {outbound_nongeo_reverse_reverse_solenoids}")

    ############################################################################
    # Flip the neccesary reference shifts
    ############################################################################
    def flip_reference_shifts(solenoid, dxy_sign, chi_sign):
        xy_shift_name   = f"{solenoid}_dxy"
        chi1_shift_name = f"{solenoid}_chi1"
        chi2_shift_name = f"{solenoid}_chi2"
        chi3_shift_name = f"{solenoid}_chi3"

        line[xy_shift_name].dx      *= dxy_sign
        line[xy_shift_name].dy      *= dxy_sign

        line[chi1_shift_name].angle *= chi_sign
        line[chi2_shift_name].angle *= chi_sign
        line[chi3_shift_name].angle *= chi_sign

    ########################################
    # Inbound Geo Forward Forward Solenoids (Complete: test_003)
    ########################################
    for inbound_geo_forward_forward_solenoid in inbound_geo_forward_forward_solenoids:
        flip_reference_shifts(
            inbound_geo_forward_forward_solenoid,
            dxy_sign    = +1,
            chi_sign    = +1)

    ########################################
    # Inbound Geo Forward Reverse Solenoids (Complete: test_003)
    ########################################
    for inbound_geo_forward_reverse_solenoid in inbound_geo_forward_reverse_solenoids:
        flip_reference_shifts(
            inbound_geo_forward_reverse_solenoid,
            dxy_sign    = +1,
            chi_sign    = +1)

    ########################################
    # Inbound Geo Reverse Forward Solenoids (Complete: test_003)
    ########################################
    for inbound_geo_reverse_forward_solenoid in inbound_geo_reverse_forward_solenoids:
        flip_reference_shifts(
            inbound_geo_reverse_forward_solenoid,
            dxy_sign    = +1,
            chi_sign    = -1)
        
    ########################################
    # Inbound Geo Reverse Reverse Solenoids (Complete: test_003)
    ########################################
    for inbound_geo_reverse_reverse_solenoid in inbound_geo_reverse_reverse_solenoids:
        flip_reference_shifts(
            inbound_geo_reverse_reverse_solenoid,
            dxy_sign    = +1,
            chi_sign    = -1)

    ########################################
    # Inbound Non-Geo Forward Forward Solenoids (Complete: test_005)
    ########################################
    for inbound_nongeo_forward_forward_solenoid in inbound_nongeo_forward_forward_solenoids:
        flip_reference_shifts(
            inbound_nongeo_forward_forward_solenoid,
            dxy_sign    = +1,
            chi_sign    = -1)
        
    ########################################
    # Inbound Non-Geo Forward Reverse Solenoids (Complete: test_005)
    ########################################
    for inbound_nongeo_forward_reverse_solenoid in inbound_nongeo_forward_reverse_solenoids:
        flip_reference_shifts(
            inbound_nongeo_forward_reverse_solenoid,
            dxy_sign    = +1,
            chi_sign    = -1)

    ########################################
    # Inbound Non-Geo Reverse Forward Solenoids (Complete: test_005)
    ########################################
    for inbound_nongeo_reverse_forward_solenoid in inbound_nongeo_reverse_forward_solenoids:
        flip_reference_shifts(
            inbound_nongeo_reverse_forward_solenoid,
            dxy_sign    = +1,
            chi_sign    = -1)

    ########################################
    # Inbound Non-Geo Reverse Reverse Solenoids (Complete: test_005)
    ########################################
    for inbound_nongeo_reverse_reverse_solenoid in inbound_nongeo_reverse_reverse_solenoids:
        flip_reference_shifts(
            inbound_nongeo_reverse_reverse_solenoid,
            dxy_sign    = +1,
            chi_sign    = -1)
        
    ########################################
    # Outbound Geo Forward Forward Solenoids (Complete: test_006)
    ########################################
    for outbound_geo_forward_forward_solenoid in outbound_geo_forward_forward_solenoids:
        flip_reference_shifts(
            outbound_geo_forward_forward_solenoid,
            dxy_sign    = +1,
            chi_sign    = -1)
        
    ########################################
    # Outbound Geo Forward Reverse Solenoids (Complete: test_006)
    ########################################
    for outbound_geo_forward_reverse_solenoid in outbound_geo_forward_reverse_solenoids:
        flip_reference_shifts(
            outbound_geo_forward_reverse_solenoid,
            dxy_sign    = +1,
            chi_sign    = +1)

    ########################################
    # Outbound Geo Reverse Forward Solenoids (Complete: test_006)
    ########################################
    for outbound_geo_reverse_forward_solenoid in outbound_geo_reverse_forward_solenoids:
        flip_reference_shifts(
            outbound_geo_reverse_forward_solenoid,
            dxy_sign    = +1,
            chi_sign    = +1)
        
    ########################################
    # Outbound Geo Reverse Reverse Solenoids (Complete: test_006)
    ########################################
    for outbound_geo_reverse_reverse_solenoid in outbound_geo_reverse_reverse_solenoids:
        flip_reference_shifts(
            outbound_geo_reverse_reverse_solenoid,
            dxy_sign    = +1,
            chi_sign    = +1)

    ########################################
    # Outbound Non-Geo Forward Forward Solenoids (Complete: test_004)
    ########################################
    for outbound_nongeo_forward_forward_solenoid in outbound_nongeo_forward_forward_solenoids:
        flip_reference_shifts(
            outbound_nongeo_forward_forward_solenoid,
            dxy_sign    = -1,
            chi_sign    = -1)

    ########################################
    # Outbound Non-Geo Forward Reverse Solenoids (Complete: test_004)
    ########################################
    for outbound_nongeo_forward_reverse_solenoid in outbound_nongeo_forward_reverse_solenoids:
        flip_reference_shifts(
            outbound_nongeo_forward_reverse_solenoid,
            dxy_sign    = +1,
            chi_sign    = -1)

    ########################################
    # Outbound Non-Geo Reverse Forward Solenoids (Complete: test_004)
    ########################################
    for outbound_nongeo_reverse_forward_solenoid in outbound_nongeo_reverse_forward_solenoids:
        flip_reference_shifts(
            outbound_nongeo_reverse_forward_solenoid,
            dxy_sign    = +1,
            chi_sign    = -1)

    ########################################
    # Outbound Non-Geo Reverse Reverse Solenoids (Complete: test_004)
    ########################################
    for outbound_nongeo_reverse_reverse_solenoid in outbound_nongeo_reverse_reverse_solenoids:
        flip_reference_shifts(
            outbound_nongeo_reverse_reverse_solenoid,
            dxy_sign    = -1,
            chi_sign    = -1)

    ############################################################################
    # Move DZ Shifts onto the outbound solenoid (even when bound)
    ############################################################################
    for inbound_solenoid, outbound_solenoid in bound_solenoid_pairs:

        # Should always be true, but just in case
        if not inbound_solenoid.endswith('_bound'):
            raise ValueError(f"Inbound solenoid {inbound_solenoid} doesn't end with bound?")
        else:
            inbound_solenoid    = inbound_solenoid[:-6]

        if not outbound_solenoid.endswith('_bound'):
            raise ValueError(f"Outbound solenoid {outbound_solenoid} doesn't end with bound?")
        else:
            outbound_solenoid   = outbound_solenoid[:-6]

    ############################################################################
    # Reorder the solenoid components
    ############################################################################
    inbound_geo_solenoids      = inbound_geo_forward_forward_solenoids + \
        inbound_geo_forward_reverse_solenoids + \
        inbound_geo_reverse_forward_solenoids + \
        inbound_geo_reverse_reverse_solenoids
    inbound_nongeo_solenoids    = inbound_nongeo_forward_forward_solenoids + \
        inbound_nongeo_forward_reverse_solenoids + \
        inbound_nongeo_reverse_forward_solenoids + \
        inbound_nongeo_reverse_reverse_solenoids
    outbound_geo_solenoids      = outbound_geo_forward_forward_solenoids + \
        outbound_geo_forward_reverse_solenoids + \
        outbound_geo_reverse_forward_solenoids + \
        outbound_geo_reverse_reverse_solenoids
    outbound_nongeo_solenoids   = outbound_nongeo_forward_forward_solenoids + \
        outbound_nongeo_forward_reverse_solenoids + \
        outbound_nongeo_reverse_forward_solenoids + \
        outbound_nongeo_reverse_reverse_solenoids
    
    inbound_geo_solenoids       = list(set(inbound_geo_solenoids))
    inbound_nongeo_solenoids    = list(set(inbound_nongeo_solenoids))
    outbound_geo_solenoids      = list(set(outbound_geo_solenoids))
    outbound_nongeo_solenoids   = list(set(outbound_nongeo_solenoids))

    ########################################
    # Get the current order of the element names
    ########################################
    element_names   = line.element_names.copy()

    ########################################
    # Reorder inbound geo solenoids
    ########################################
    for inbound_geo_solenoid in tqdm(inbound_geo_solenoids):
        
        sol_start_ele   = f"{inbound_geo_solenoid}_bound"
        sol_end_ele     = f"{inbound_geo_solenoid}_chi3"
        
        # Get the start and end indices
        start_idxs  = [i for i, name in enumerate(element_names) if name == sol_start_ele]
        end_idxs    = [i for i, name in enumerate(element_names) if name == sol_end_ele]

        for start_idx, end_idx in zip(start_idxs, end_idxs):
            assert start_idx < end_idx

            new_element_names   = []
            new_element_names   += element_names[:start_idx]
            bound_elements      = [
                f'{inbound_geo_solenoid}_chi3',
                f'{inbound_geo_solenoid}_chi2',
                f'{inbound_geo_solenoid}_chi1',
                f'{inbound_geo_solenoid}_dz',
                f'{inbound_geo_solenoid}_dxy',
                f'{inbound_geo_solenoid}_bound']
            new_element_names   += bound_elements
            new_element_names   += element_names[end_idx + 1:]

            element_names       = new_element_names

    ########################################
    # Reorder inbound non-geo solenoids
    ########################################
    for inbound_nongeo_solenoid in tqdm(inbound_nongeo_solenoids):

        sol_start_ele   = f"{inbound_nongeo_solenoid}_bound"
        sol_end_ele     = f"{inbound_nongeo_solenoid}_chi3"
        
        # Get the start and end indices
        start_idxs  = [i for i, name in enumerate(element_names) if name == sol_start_ele]
        end_idxs    = [i for i, name in enumerate(element_names) if name == sol_end_ele]

        for start_idx, end_idx in zip(start_idxs, end_idxs):
            assert start_idx < end_idx

            new_element_names   = []
            new_element_names   += element_names[:start_idx]
            if not reverse_line:
                bound_elements      = [
                    f'{inbound_nongeo_solenoid}_chi1',
                    f'{inbound_nongeo_solenoid}_chi2',
                    f'{inbound_nongeo_solenoid}_chi3',
                    f'{inbound_nongeo_solenoid}_dz',
                    f'{inbound_nongeo_solenoid}_dxy',
                    f'{inbound_nongeo_solenoid}_bound']
            else:
                bound_elements      = [
                    f'{inbound_nongeo_solenoid}_chi3',
                    f'{inbound_nongeo_solenoid}_chi2',
                    f'{inbound_nongeo_solenoid}_chi1',
                    f'{inbound_nongeo_solenoid}_dz',
                    f'{inbound_nongeo_solenoid}_dxy',
                    f'{inbound_nongeo_solenoid}_bound']
            new_element_names   += bound_elements
            new_element_names   += element_names[end_idx + 1:]
            
            element_names       = new_element_names

    ########################################
    # Reorder outbound geo solenoids
    ########################################
    for outbound_geo_solenoid in tqdm(outbound_geo_solenoids):

        sol_start_ele   = f"{outbound_geo_solenoid}_bound"
        sol_end_ele     = f"{outbound_geo_solenoid}_chi3"
        
        # Get the start and end indices
        start_idxs  = [i for i, name in enumerate(element_names) if name == sol_start_ele]
        end_idxs    = [i for i, name in enumerate(element_names) if name == sol_end_ele]

        for start_idx, end_idx in zip(start_idxs, end_idxs):
            assert start_idx < end_idx

            new_element_names   = []
            new_element_names   += element_names[:start_idx]
            bound_elements      = [
                f'{outbound_geo_solenoid}_bound',
                f'{outbound_geo_solenoid}_dxy',
                f'{outbound_geo_solenoid}_dz',
                f'{outbound_geo_solenoid}_chi1',
                f'{outbound_geo_solenoid}_chi2',
                f'{outbound_geo_solenoid}_chi3']
            new_element_names   += bound_elements
            new_element_names   += element_names[end_idx + 1:]
            
            element_names       = new_element_names

    ########################################
    # Reorder outbound non-geo solenoids
    ########################################
    for outbound_nongeo_solenoid in tqdm(outbound_nongeo_solenoids):

        sol_start_ele   = f"{outbound_nongeo_solenoid}_bound"
        sol_end_ele     = f"{outbound_nongeo_solenoid}_chi3"
        
        # Get the start and end indices
        start_idxs  = [i for i, name in enumerate(element_names) if name == sol_start_ele]
        end_idxs    = [i for i, name in enumerate(element_names) if name == sol_end_ele]

        for start_idx, end_idx in zip(start_idxs, end_idxs):
            assert start_idx < end_idx
            
            new_element_names   = []
            new_element_names   += element_names[:start_idx]
            bound_elements      = [
                f'{outbound_nongeo_solenoid}_bound',
                f'{outbound_nongeo_solenoid}_dxy',
                f'{outbound_nongeo_solenoid}_dz',
                f'{outbound_nongeo_solenoid}_chi1',
                f'{outbound_nongeo_solenoid}_chi2',
                f'{outbound_nongeo_solenoid}_chi3']
            new_element_names   += bound_elements
            new_element_names   += element_names[end_idx + 1:]
            element_names       = new_element_names

    ########################################
    # Update the line
    ########################################
    line.element_names = element_names



"""
(Unofficial) SAD to XSuite Converter
"""

################################################################################
# Import Packages
################################################################################
from datetime import date

from ..output_writer._001_drift import create_drift_lattice_file_information
from ..output_writer._002_bend import create_bend_lattice_file_information
from ..output_writer._003_corr import create_corrector_lattice_file_information
from ..output_writer._004_quad import create_quadrupole_lattice_file_information
from ..output_writer._005_sext import create_sextupole_lattice_file_information
from ..output_writer._006_oct import create_octupole_lattice_file_information
from ..output_writer._007_mult import create_multipole_lattice_file_information
from ..output_writer._008_sol import create_solenoid_lattice_file_information
from ..output_writer._009_cavity import create_cavity_lattice_file_information
from ..output_writer._010_refshift import create_refshift_lattice_file_information
from ..output_writer._011_marker import create_marker_lattice_file_information
from ..output_writer._012_line import create_line_lattice_file_information
from ..output_writer._013_model import create_model_lattice_file_information
from ..output_writer._014_offset_markers import create_offset_marker_lattice_file_information

today   = date.today()

################################################################################
# Write the lattice file
################################################################################
def write_lattice(
        line,
        offset_marker_locations,
        output_filename,
        output_directory,
        output_header):
    """
    Write the outputs to the specified files.
    
    Parameters:
    line (xt.Line): The xtrack line object.
    output_filename (str): The base name for the output files.
    header (str): The header for the output files.
    """

    ########################################
    # Initialise the lattice file
    ########################################
    lattice_file_string = f'''"""
{output_header}
================================================================================
Converted using the SAD2XS Converter
Authors:    J. Salvesen
Contact:    john.salvesen@cern.ch
================================================================================
Conversion Date: {today.strftime("%d/%m/%Y")}
"""

################################################################################
# Import Packages
################################################################################
import xtrack as xt
import numpy as np

################################################################################
# Create or Get Environment
################################################################################
env = xt.get_environment(verbose = True)
env.vars.default_to_zero = True

########################################
# Key Global Variables
########################################
env["mass"]     = {line.particle_ref.mass0}
env["p0c"]      = {line.particle_ref.p0c[0]}
env["q0"]       = {line.particle_ref.q0}
env["fshift"]   = {line.env["fshift"]}

########################################
# Reference Particle
########################################
env.particle_ref    = xt.Particles(
    mass0   = env["mass"],
    p0c     = env["p0c"],
    q0      = env["q0"])

################################################################################
# Import lattice
################################################################################
'''
    
    ########################################
    # Add all the other sections
    ########################################
    line_table  = line.get_table(attr = True)

    lattice_file_string += create_drift_lattice_file_information(line, line_table)
    lattice_file_string += create_bend_lattice_file_information(line, line_table)
    lattice_file_string += create_corrector_lattice_file_information(line, line_table)
    lattice_file_string += create_quadrupole_lattice_file_information(line, line_table)
    lattice_file_string += create_sextupole_lattice_file_information(line, line_table)
    lattice_file_string += create_octupole_lattice_file_information(line, line_table)
    lattice_file_string += create_multipole_lattice_file_information(line, line_table)
    lattice_file_string += create_solenoid_lattice_file_information(line, line_table)
    lattice_file_string += create_cavity_lattice_file_information(line, line_table)
    lattice_file_string += create_refshift_lattice_file_information(line, line_table)
    lattice_file_string += create_marker_lattice_file_information(line, line_table, offset_marker_locations)
    lattice_file_string += create_line_lattice_file_information(line, line_table)
    lattice_file_string += create_model_lattice_file_information()
    lattice_file_string += create_offset_marker_lattice_file_information(offset_marker_locations)

    ########################################
    # Write to file
    ########################################
    with open(f"{output_directory}/{output_filename}.py", 'w') as f:
        f.write(lattice_file_string)



"""
(Unofficial) SAD to XSuite Converter
"""

################################################################################
# Import Packages
################################################################################
import numpy as np
from .._globals import *

################################################################################
# Naming
################################################################################

########################################
# Parent/Variable Name Extraction
########################################
def get_parentname(element_name):
    """
    Elements may be repeated or inverted
    Repeated elements are suffixed with ::0, ::1 etc
    """
    # Assume to start that the parent name is the element name excluding replica
    parent_name    = element_name.split('::')[0]
    
    return parent_name

def get_variablename(element_name):
    """
    Elements may be repeated or inverted
    Repeated elements are suffixed with ::0, ::1 etc
    Inverted elements are prefixed with a '-' but take the same variable as the non-inverted element
    """

    # Get the parent name
    parent_name     = get_parentname(element_name)

    # If the element is inverted, the variable name needs the '-' removed
    if parent_name.startswith('-'):
        variable_name   = parent_name[1:]
    else:
        variable_name   = parent_name
    
    return variable_name

################################################################################
# Elements for replication naming
################################################################################
def generate_magnet_for_replication_names(length_dict, base_string):
    """
    To simplify the output, we create base magnets for replication
    This function generates the names for these base magnets
    Convention based on magnet type and length in microns

    N.B. Two assumptions:
    1. Lengths are rounded to the nearest micron
    2. Lengths are less than 10m
    3. Lengths are non-negative
    """
    names           = []
    length_values	= np.array(list(length_dict.keys()))
    length_values	= length_values * 1E9
    length_values	= length_values.astype(int)
    
    for length in length_values:
        name = f"{base_string}{length:011d}"
        names.append(name)
    names = sorted(names)
    
    return names

################################################################################
# KNL/KSL arrays to strings
################################################################################
def get_knl_string(knl_array):
    """
    Returns a string with the knl values
    """
    # If all zero, just give an empty array
    if np.all(knl_array == 0):
        return "[]"
    
    # Otherwise, iterate through
    knl_string = "["
    for i, knl in enumerate(knl_array):

        # If there are no more knl values, close:
        if np.all(knl_array[i:] == 0):
            break

        # Fromat the knl value
        if knl == 0:
            knl_substring   = f"0"
        else:
            knl_substring   = f"{knl:.10e}"
        
        # Append to the string
        if i == 0:
            knl_string += knl_substring
        else:
            knl_string += f', {knl_substring}'
    
    # Close the string
    knl_string += "]"
    return knl_string

################################################################################
# Extract Magnet Information
################################################################################

########################################
# Bends
########################################
def extract_bend_information(line, line_table):

    ########################################
    # Get Bend Element information
    ########################################
    unique_bend_names           = []
    unique_bend_variables       = []

    for bend in line_table.rows[line_table.element_type == 'Bend'].name:
        parentname      = get_parentname(bend)
        variablename    = get_variablename(bend)
        
        # Ensure the element is a bend not a corrector
        if line[parentname].h != 0:
            if parentname not in unique_bend_names:
                unique_bend_names.append(parentname)
                unique_bend_variables.append(variablename)

    bend_name_dict      = {}
    for bend_name, bend_variable in zip(unique_bend_names, unique_bend_variables):
        bend_name_dict[bend_name] = bend_variable

    unique_bend_variables       = sorted(list(set(unique_bend_variables)))

    ########################################
    # Bend Base Element information
    ########################################
    # Get the base elements to replicate: pure horizontal, vertical and skew
    hbends	= {}
    vbends  = {}
    sbends  = {}

    for bend in unique_bend_names:

        # Get the length and rotation of the bend
        length		= line[bend].length
        rot_s_rad	= line[bend].rot_s_rad

        ########################################
        # Categorise H and V based on the rotation
        ########################################
        # Mapping from rotation → (target‑dict, need_flip)
        angle_map = {
            0:               (hbends, False),
            np.pi:           (hbends, True),
            -np.pi / 2:      (vbends, True),
            np.pi / 2:       (vbends, False)}

        # Try to match one of the valid angles
        angle_matched   = False
        for angle, (bend_dict, flip) in angle_map.items():
            if np.isclose(rot_s_rad, angle):
                angle_matched   = True
                
                # Handle horizontal and vertical bends rotated by 180 degrees
                if flip:
                    line[bend].k0               *= -1
                    line[bend].h                *= -1
                    line[bend].edge_entry_angle *= -1
                    line[bend].edge_exit_angle  *= -1
                    line[bend].rot_s_rad        *= -1

                # insert without duplicates
                lst = bend_dict.setdefault(length, [])
                if bend not in lst:
                    lst.append(bend)
                # else: already there → skip silently
                break

        if not angle_matched:
            # → skew bend
            lst = sbends.setdefault(length, [])
            if bend not in lst:
                lst.append(bend)

    return hbends, vbends, sbends, unique_bend_variables, bend_name_dict

########################################
# Correctors
########################################
def extract_corrector_information(line, line_table):

    ########################################
    # Get Corrector Element information
    ########################################
    unique_corr_names           = []
    unique_corr_variables       = []

    for corr in line_table.rows[line_table.element_type == 'Bend'].name:
        parentname      = get_parentname(corr)
        variablename    = get_variablename(corr)
        
        # Ensure the element is a corr not a corrector
        if line[parentname].h == 0:
            if parentname not in unique_corr_names:
                unique_corr_names.append(parentname)
                unique_corr_variables.append(variablename)

    corr_name_dict      = {}
    for corr_name, corr_variable in zip(unique_corr_names, unique_corr_variables):
        corr_name_dict[corr_name] = corr_variable

    unique_corr_variables       = sorted(list(set(unique_corr_variables)))

    ########################################
    # corr Base Element information
    ########################################
    # Get the base elements to replicate: pure horizontal, vertical and skew
    hcorrs	= {}
    vcorrs  = {}
    scorrs  = {}

    for corr in unique_corr_names:

        # Get the length and rotation of the corr
        length		= line[corr].length
        rot_s_rad	= line[corr].rot_s_rad

        ########################################
        # Categorise H and V based on the rotation
        ########################################
        # Mapping from rotation → (target‑dict, need_flip)
        angle_map = {
            0:               (hcorrs, False),
            np.pi:           (hcorrs, True),
            -np.pi / 2:      (vcorrs, True),
            np.pi / 2:       (vcorrs, False)}

        # Try to match one of the valid angles
        angle_matched   = False
        for angle, (corr_dict, flip) in angle_map.items():
            if np.isclose(rot_s_rad, angle):
                angle_matched   = True

                # Handle horizontal and vertical corrs rotated by 180 degrees
                if flip:
                    line[corr].k0               *= -1
                    line[corr].h                *= -1
                    line[corr].edge_entry_angle *= -1
                    line[corr].edge_exit_angle  *= -1
                    line[corr].rot_s_rad        *= -1

                # insert without duplicates
                lst = corr_dict.setdefault(length, [])
                if corr not in lst:
                    lst.append(corr)
                # else: already there → skip silently
                break

        if not angle_matched:
            # → skew corr
            lst = scorrs.setdefault(length, [])
            if corr not in lst:
                lst.append(corr)

    return hcorrs, vcorrs, scorrs, unique_corr_variables, corr_name_dict

########################################
# Quadrupole/Sextupole/Octupole information
########################################
def extract_multipole_information(line, line_table, mode):

    ########################################
    # Get Magnet Element information
    ########################################
    unique_names       = []
    for magnet in line_table.rows[line_table.element_type == mode].name:
        parentname      = get_parentname(magnet)
        if parentname not in unique_names:
            unique_names.append(parentname)

    ########################################
    # Magnets based on length
    ########################################
    magnets   = {}
    for magnet in unique_names:
        length		= line[magnet].length
        if length not in magnets:
            magnets[length] = [magnet]
        else:
            if magnet not in magnets[length]:
                magnets[length].append(magnet)
            else:
                continue

    return magnets, unique_names

################################################################################
# Element is simple to clone
################################################################################
def check_is_simple_bend_corr(line, replica_name):
    is_simple = False
    
    if line[replica_name].edge_entry_angle == 0 and \
            line[replica_name].edge_exit_angle == 0 and \
            line[replica_name].edge_entry_angle_fdown == 0 and \
            line[replica_name].edge_exit_angle_fdown == 0 and \
            line[replica_name].shift_x == 0 and \
            line[replica_name].shift_y == 0:
        is_simple = True
    
    return is_simple

def check_is_simple_quad_sext_oct(line, replica_name, mode):
    is_simple   = False
    is_skew     = False

    if mode == "Quadrupole":
        # Simple assumes only one of k1 or k1s is non-zero
        if (line[replica_name].k1 * line[replica_name].k1s) == 0 and \
                line[replica_name].shift_x == 0 and \
                line[replica_name].shift_y == 0 and \
                line[replica_name].rot_s_rad == 0:
            is_simple = True

    if mode == "Sextupole":
        # Simple assumes only one of k2 or k2s is non-zero
        if (line[replica_name].k2 * line[replica_name].k2s) == 0 and \
                line[replica_name].shift_x == 0 and \
                line[replica_name].shift_y == 0 and \
                line[replica_name].rot_s_rad == 0:
            is_simple = True

    if mode == "Octupole":
        # Simple assumes only one of k3 or k3s is non-zero
        if (line[replica_name].k3 * line[replica_name].k3s) == 0 and \
                line[replica_name].shift_x == 0 and \
                line[replica_name].shift_y == 0 and \
                line[replica_name].rot_s_rad == 0:
            is_simple = True

    return is_simple

def check_is_skew_quad_sext_oct(line, replica_name, mode):
    is_skew     = False

    if mode == "Quadrupole":
        if line[replica_name].k1s != 0:
            is_skew = True

    if mode == "Sextupole":
        if line[replica_name].k2s != 0:
            is_skew = True

    if mode == "Octupole":
        if line[replica_name].k3s != 0:
            is_skew = True

    return is_skew

def check_is_simple_unpowered_multipole(line, replica_name):
    is_simple_unpowered = False

    if np.all(line[replica_name].knl == 0) and \
            np.all(line[replica_name].ksl == 0) and \
            line[replica_name].shift_x == 0 and \
            line[replica_name].shift_y == 0 and \
            line[replica_name].rot_s_rad == 0:
        is_simple_unpowered = True

    return is_simple_unpowered

def check_is_simple_solenoid(line, replica_name):
    is_simple_unpowered = False

    if np.all(line[replica_name].knl == 0) and \
            np.all(line[replica_name].ksl == 0) and \
            line[replica_name].mult_shift_x == 0 and \
            line[replica_name].mult_shift_y == 0 and \
            line[replica_name].rot_s_rad == 0:
        is_simple_unpowered = True

    return is_simple_unpowered



"""
(Unofficial) SAD to XSuite Converter
"""

################################################################################
# Required Packages
################################################################################
import os
import sad2xs as s2x
import xtrack as xt
import numpy as np
import textwrap
import matplotlib.pyplot as plt
from tqdm import tqdm

from _sad_helpers import sad_twiss, rebuild_sad_lattice
from _test_config import *

################################################################################
# Reference PyTest Function
################################################################################
def reference_sol_test(
        test_name:                  str,
        sad_elements_line_string:   str,
        sad_eval_marker:            str         = "END",
        test_values:                np.ndarray  = TEST_VALUES,
        static_val:                 float       = STATIC_OFFSET,
        plot:                       bool        = True):
    """
    Reference Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """

    ############################################################################
    # Scan
    ############################################################################
    S_SAD       = np.zeros_like(test_values)
    S_XS        = np.zeros_like(test_values)
    X_SAD       = np.zeros_like(test_values)
    X_XS        = np.zeros_like(test_values)
    Y_SAD       = np.zeros_like(test_values)
    Y_XS        = np.zeros_like(test_values)
    PX_SAD      = np.zeros_like(test_values)
    PX_XS       = np.zeros_like(test_values)
    PY_SAD      = np.zeros_like(test_values)
    PY_XS       = np.zeros_like(test_values)

    for iteration, test_val in enumerate(tqdm(test_values)):

        ########################################################################
        # Write Test Lattice
        ########################################################################
        with open("test_lattice.sad", "w") as f:
            f.write(textwrap.dedent(f"""\
            MOMENTUM    = 1.0 GEV;
            TEST_VAL    = {test_val};
            STATIC_VAL  = {static_val};
            """))

        with open("test_lattice.sad", "a") as f:
            f.write(sad_elements_line_string)

        ########################################################################
        # Twiss SAD Lattice
        ########################################################################
        tw_sad  = sad_twiss(
            lattice_filename    = 'test_lattice.sad',
            line_name           = 'TEST_LINE')
        
        ########################################################################
        # Rebuild SAD Lattice
        ########################################################################
        rebuild_sad_lattice(
            lattice_filename    = 'test_lattice.sad',
            line_name           = 'TEST_LINE')

        ########################################################################
        # Convert Lattice
        ########################################################################
        line    = s2x.convert_sad_to_xsuite(
            sad_lattice_path    = 'test_lattice.sad',
            output_directory    = None,
            _verbose            = False,
            _test_mode          = True)

        ########################################
        # Get table XSuite Lattice
        ########################################
        tt      = line.get_table(attr = True)

        ########################################
        # Twiss XSuite Lattice
        ########################################
        tw_xs   = line.twiss4d(
            _continue_if_lost   = True,
            start               = xt.START,
            end                 = xt.END,
            betx                = 1,
            bety                = 1)
        
        ########################################
        # Calculate SAD s
        ########################################
        ds              = np.concatenate([[0], tw_xs.s[1:] - tw_xs.s[:-1]])
        dzeta           = np.concatenate([[0], tw_xs.zeta[1:] - tw_xs.zeta[:-1]])
        # Ignore the dzeta at ZetaShift elements
        zeta_shifts     = tt.element_type == "ZetaShift"
        zeta_shifts     = np.concatenate([[0], zeta_shifts[:-1]])
        dzeta           = np.where(zeta_shifts, 0, dzeta)

        s_sad           = np.zeros_like(tw_xs.s)
        for i in range(1, len(s_sad)):
            s_sad[i]    = s_sad[i-1] + ds[i] - dzeta[i]
        tw_xs["s_sad"]     = s_sad

        ########################################################################
        # Save data for comparison
        ########################################################################
        xs_eval_marker      = sad_eval_marker.lower()

        S_SAD[iteration]    = tw_sad["s", sad_eval_marker]
        S_XS[iteration]     = tw_xs["s_sad", xs_eval_marker]
        X_SAD[iteration]    = tw_sad["x", sad_eval_marker]
        X_XS[iteration]     = tw_xs["x", xs_eval_marker]
        Y_SAD[iteration]    = tw_sad["y", sad_eval_marker]
        Y_XS[iteration]     = tw_xs["y", xs_eval_marker]
        PX_SAD[iteration]   = tw_sad["px", sad_eval_marker]
        PX_XS[iteration]    = tw_xs["px", xs_eval_marker]
        PY_SAD[iteration]   = tw_sad["py", sad_eval_marker]
        PY_XS[iteration]    = tw_xs["py", xs_eval_marker]

        ########################################################################
        # Delete test lattice
        ########################################################################
        os.remove("test_lattice.sad")

        ########################################################################
        # Diagnostics: Outputs if the test fails
        ########################################################################
        if not (np.isclose(S_SAD[iteration], S_XS[iteration], rtol = DELTA_S_RTOL, atol = DELTA_S_ATOL) and \
                np.isclose(X_SAD[iteration], X_XS[iteration], rtol = DELTA_X_RTOL, atol = DELTA_X_ATOL) and \
                np.isclose(Y_SAD[iteration], Y_XS[iteration], rtol = DELTA_Y_RTOL, atol = DELTA_Y_ATOL) and \
                np.isclose(PX_SAD[iteration], PX_XS[iteration], rtol = DELTA_PX_RTOL, atol = DELTA_PX_ATOL) and \
                np.isclose(PY_SAD[iteration], PY_XS[iteration], rtol = DELTA_PY_RTOL, atol = DELTA_PY_ATOL)):

            print("SAD")
            tw_sad.cols["s x px y py"].show()
            print("Xsuite")
            tw_xs.cols["s_sad x px y py"].show()

            delta_s_in_pre      = (tw_xs["s_sad", "start"] - tw_sad["s", "START"]) * 1E9
            delta_s_in_post     = (tw_xs["s_sad", "sol_start"] - tw_sad["s", "SOL_START"]) * 1E9
            delta_x_in_pre      = (tw_xs["x", "start"] - tw_sad["x", "START"]) * 1E9
            delta_x_in_post     = (tw_xs["x", "sol_start"] - tw_sad["x", "SOL_START"]) * 1E9
            delta_y_in_pre      = (tw_xs["y", "start"] - tw_sad["y", "START"]) * 1E9
            delta_y_in_post     = (tw_xs["y", "sol_start"] - tw_sad["y", "SOL_START"]) * 1E9
            delta_px_in_pre     = (tw_xs["px", "start"] - tw_sad["px", "START"]) * 1E9
            delta_px_in_post    = (tw_xs["px", "sol_start"] - tw_sad["px", "SOL_START"]) * 1E9
            delta_py_in_pre     = (tw_xs["py", "start"] - tw_sad["py", "START"]) * 1E9
            delta_py_in_post    = (tw_xs["py", "sol_start"] - tw_sad["py", "SOL_START"]) * 1E9

            delta_s_out_pre     = (tw_xs["s_sad", "sol_end"] - tw_sad["s", "SOL_END"]) * 1E9
            delta_s_out_post    = (tw_xs["s_sad", "end"] - tw_sad["s", "END"]) * 1E9
            delta_x_out_pre     = (tw_xs["x", "sol_end"] - tw_sad["x", "SOL_END"]) * 1E9
            delta_x_out_post    = (tw_xs["x", "end"] - tw_sad["x", "END"]) * 1E9
            delta_y_out_pre     = (tw_xs["y", "sol_end"] - tw_sad["y", "SOL_END"]) * 1E9
            delta_y_out_post    = (tw_xs["y", "end"] - tw_sad["y", "END"]) * 1E9
            delta_px_out_pre    = (tw_xs["px", "sol_end"] - tw_sad["px", "SOL_END"]) * 1E9
            delta_px_out_post   = (tw_xs["px", "end"] - tw_sad["px", "END"]) * 1E9
            delta_py_out_pre    = (tw_xs["py", "sol_end"] - tw_sad["py", "SOL_END"]) * 1E9
            delta_py_out_post   = (tw_xs["py", "end"] - tw_sad["py", "END"]) * 1E9

            print("Inbound Before transforms (element START)")
            print(f"Xsuite: s  = {tw_xs["s_sad", "start"]:.3f}, x = {tw_xs["x", "start"]:.3f}, y = {tw_xs["y", "start"]:.3f}, px = {tw_xs["px", "start"]:.3f}, py = {tw_xs["py", "start"]:.3f}")
            print(f"SAD:    s  = {tw_sad["s", "START"]:.3f}, x = {tw_sad["x", "START"]:.3f}, y = {tw_sad["y", "START"]:.3f}, px = {tw_sad["px", "START"]:.3f}, py = {tw_sad["py", "START"]:.3f}")
            print(f"Delta:  s = {delta_s_in_pre:.3f}, x = {delta_x_in_pre:.3f}, y = {delta_y_in_pre:.3f}, px = {delta_px_in_pre:.3f}, py = {delta_py_in_pre:.3f}")

            print("Inbound Post transforms (element SOL_START)")
            print(f"Xsuite: s  = {tw_xs["s_sad", "sol_start"]:.3f}, x = {tw_xs["x", "sol_start"]:.3f}, y = {tw_xs["y", "sol_start"]:.3f}, px = {tw_xs["px", "sol_start"]:.3f}, py = {tw_xs["py", "sol_start"]:.3f}")
            print(f"SAD:    s  = {tw_sad["s", "SOL_START"]:.3f}, x = {tw_sad["x", "SOL_START"]:.3f}, y = {tw_sad["y", "SOL_START"]:.3f}, px = {tw_sad["px", "SOL_START"]:.3f}, py = {tw_sad["py", "SOL_START"]:.3f}")
            print(f"Delta:  s = {delta_s_in_post:.3f}, x = {delta_x_in_post:.3f}, y = {delta_y_in_post:.3f}, px = {delta_px_in_post:.3f}, py = {delta_py_in_post:.3f}")

            print("Outbound Pre transforms (element SOL_END)")
            print(f"Xsuite: s  = {tw_xs["s_sad", "sol_end"]:.3f}, x = {tw_xs["x", "sol_end"]:.3f}, y = {tw_xs["y", "sol_end"]:.3f}, px = {tw_xs["px", "sol_end"]:.3f}, py = {tw_xs["py", "sol_end"]:.3f}")
            print(f"SAD:    s  = {tw_sad["s", "SOL_END"]:.3f}, x = {tw_sad["x", "SOL_END"]:.3f}, y = {tw_sad["y", "SOL_END"]:.3f}, px = {tw_sad["px", "SOL_END"]:.3f}, py = {tw_sad["py", "SOL_END"]:.3f}")
            print(f"Delta:  s = {delta_s_out_pre:.3f}, x = {delta_x_out_pre:.3f}, y = {delta_y_out_pre:.3f}, px = {delta_px_out_pre:.3f}, py = {delta_py_out_pre:.3f}")

            print("Outbound Post transforms (element END)")
            print(f"Xsuite: s  = {tw_xs["s_sad", "end"]:.3f}, x = {tw_xs["x", "end"]:.3f}, y = {tw_xs["y", "end"]:.3f}, px = {tw_xs["px", "end"]:.3f}, py = {tw_xs["py", "end"]:.3f}")
            print(f"SAD:    s  = {tw_sad["s", "END"]:.3f}, x = {tw_sad["x", "END"]:.3f}, y = {tw_sad["y", "END"]:.3f}, px = {tw_sad["px", "END"]:.3f}, py = {tw_sad["py", "END"]:.3f}")
            print(f"Delta:  s = {delta_s_out_post:.3f}, x = {delta_x_out_post:.3f}, y = {delta_y_out_post:.3f}, px = {delta_px_out_post:.3f}, py = {delta_py_out_post:.3f}")

            print("Transform sizes Xsuite")
            print("In:")
            print(f"dx = {line['sol_in_dxy'].dx}, dy = {line['sol_in_dxy'].dy}, dz = {line['sol_in_dz'].dzeta}")
            print(f"chi1 = {line['sol_in_chi1'].angle}, chi2 = {line['sol_in_chi2'].angle}, chi3 = {line['sol_in_chi3'].angle}")
            print("Out:")
            print(f"dx = {line['sol_out_dxy'].dx}, dy = {line['sol_out_dxy'].dy}, dz = {line['sol_out_dz'].dzeta}")
            print(f"chi1 = {line['sol_out_chi1'].angle}, chi2 = {line['sol_out_chi2'].angle}, chi3 = {line['sol_out_chi3'].angle}")

    ############################################################################
    # Plot
    ############################################################################
    if plot:

        symlog_threshold    = np.nanmin(
            np.where(TEST_VALUES > 0, TEST_VALUES, np.nan))

        if not np.all(np.isclose(S_SAD, S_XS, rtol = DELTA_S_RTOL, atol = DELTA_S_ATOL)):
            fig, axs    = plt.subplots(
                nrows       = 2,
                ncols       = 1,
                figsize     = (12, 6),
            sharex      = True,
            gridspec_kw = {'height_ratios': [3, 1]})

            axs[0].plot(TEST_VALUES, S_SAD, label = "SAD", color = "r")
            axs[0].plot(TEST_VALUES, S_XS, label = "XS", color = "b", linestyle='--')
            axs[1].plot(TEST_VALUES, S_XS - S_SAD)

            axs[1].set_xlabel("Drift Length [m]")
            axs[0].set_ylabel('s [m]')
            axs[1].set_ylabel('∆s [m]')

            for ax in axs:
                ax.set_xscale('symlog', linthresh = symlog_threshold)
                ax.set_yscale('symlog', linthresh = symlog_threshold)
                ax.grid()
            axs[0].legend()
            plt.subplots_adjust(hspace = 0)
            plt.savefig(f"outputs/{test_name}_s.png", dpi = 300, bbox_inches = 'tight')

        if not (
                np.all(np.isclose(X_SAD, X_XS, rtol = DELTA_X_RTOL, atol = DELTA_X_ATOL)) and
                np.all(np.isclose(Y_SAD, Y_XS, rtol = DELTA_Y_RTOL, atol = DELTA_Y_ATOL))):
            fig, axs    = plt.subplots(
                nrows       = 2,
                ncols       = 2,
                figsize     = (12, 6),
                sharex      = True,
            gridspec_kw = {'height_ratios': [3, 1]})

            axs[0, 0].plot(TEST_VALUES, X_SAD, label = "SAD", color = "r")
            axs[0, 0].plot(TEST_VALUES, X_XS, label = "XS", color = "b", linestyle='--')
            axs[1, 0].plot(TEST_VALUES, X_XS - X_SAD)

            axs[0, 1].plot(TEST_VALUES, Y_SAD, label = "SAD", color = "r")
            axs[0, 1].plot(TEST_VALUES, Y_XS, label = "XS", color = "b", linestyle='--')
            axs[1, 1].plot(TEST_VALUES, Y_XS - Y_SAD)

            axs[1, 0].set_xlabel("Transform Parameter")
            axs[1, 1].set_xlabel("Transform Parameter")
            axs[0, 0].set_ylabel('x [m]')
            axs[1, 0].set_ylabel('∆x [m]')
            axs[0, 1].set_ylabel('y [m]')
            axs[1, 1].set_ylabel('∆y [m]')

            for ax in axs.flatten():
                ax.set_xscale('symlog', linthresh = symlog_threshold)
                ax.set_yscale('symlog', linthresh = symlog_threshold)
                ax.grid()
            axs[0, 0].legend()
            axs[0, 1].legend()
            plt.subplots_adjust(hspace = 0)
            plt.savefig(f"outputs/{test_name}_xy.png", dpi = 300, bbox_inches = 'tight')

        if not (
                np.all(np.isclose(PX_SAD, PX_XS, rtol = DELTA_X_RTOL, atol = DELTA_X_ATOL)) and
                np.all(np.isclose(PY_SAD, PY_XS, rtol = DELTA_Y_RTOL, atol = DELTA_Y_ATOL))):
            fig, axs    = plt.subplots(
                nrows       = 2,
                ncols       = 2,
                figsize     = (12, 6),
                sharex      = True,
                gridspec_kw = {'height_ratios': [3, 1]})

            axs[0, 0].plot(TEST_VALUES, PX_SAD, label = "SAD", color = "r")
            axs[0, 0].plot(TEST_VALUES, PX_XS, label = "XS", color = "b", linestyle='--')
            axs[1, 0].plot(TEST_VALUES, PX_XS - PX_SAD)

            axs[0, 1].plot(TEST_VALUES, PY_SAD, label = "SAD", color = "r")
            axs[0, 1].plot(TEST_VALUES, PY_XS, label = "XS", color = "b", linestyle='--')
            axs[1, 1].plot(TEST_VALUES, PY_XS - PY_SAD)

            axs[1, 0].set_xlabel("Transform Parameter")
            axs[1, 1].set_xlabel("Transform Parameter")
            axs[0, 0].set_ylabel('px [m]')
            axs[1, 0].set_ylabel('∆px [m]')
            axs[0, 1].set_ylabel('py [m]')
            axs[1, 1].set_ylabel('∆py [m]')

            for ax in axs.flatten():
                ax.set_xscale('symlog', linthresh = symlog_threshold)
                ax.set_yscale('symlog', linthresh = symlog_threshold)
                ax.grid()
            axs[0, 0].legend()
            axs[0, 1].legend()
            plt.subplots_adjust(hspace = 0)
            plt.savefig(f"outputs/{test_name}_pxpy.png", dpi = 300, bbox_inches = 'tight')

        plt.close("all")

    ############################################################################
    # Assertions
    ############################################################################
    assert np.all(
        np.isclose(S_SAD, S_XS, rtol = DELTA_S_RTOL, atol = DELTA_S_ATOL)), \
        "s values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(X_SAD, X_XS, rtol = DELTA_X_RTOL, atol = DELTA_X_ATOL)), \
        "x values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(Y_SAD, Y_XS, rtol = DELTA_Y_RTOL, atol = DELTA_Y_ATOL)), \
        "y values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(PX_SAD, PX_XS, rtol = DELTA_PX_RTOL, atol = DELTA_PX_ATOL)), \
        "px values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(PY_SAD, PY_XS, rtol = DELTA_PY_RTOL, atol = DELTA_PY_ATOL)), \
        "py values do not match between SAD and XSuite."

# ################################################################################
# # Very weak sol (checking expansion of the square root)
# ################################################################################
# def test_sol_on_very_weak():
#     """
#     Test the conversion of a SAD SOL element with ref shifts at entry to XSuite.
#     """
#     reference_sol_test(
#         test_name                  = "test_007_very_weak_sol",
#         sad_elements_line_string   = textwrap.dedent(f"""\
#             DRIFT       SHORT_DRIFT = (L = 0.10);

#             SOL         SOL_IN      = (BZ = 0.000001 BOUND =1  DX = TEST_VAL DY = TEST_VAL CHI1 = TEST_VAL CHI2 = TEST_VAL GEO = 1)
#                         SOL_OUT     = (BZ = 0.000001 BOUND =1);

#             MARK        START       = ()
#                         END         = ()
#                         SOL_START   = ()
#                         SOL_END     = ();

#             LINE        SOL_DRIFT   = (SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT
#                 SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT
#                 SHORT_DRIFT SHORT_DRIFT)
#                         TEST_LINE   = (START
#                 SOL_IN SOL_START SOL_DRIFT SOL_END SOL_OUT END);
#             """),
#         sad_eval_marker            = "END",
#         test_values                = generate_symlog_array(-6, -1, 11),
#         static_val                 = STATIC_OFFSET,
#         plot                       = True)

# ################################################################################
# # Weak sol (checking expansion of the square root)
# ################################################################################
# def test_sol_on_weak():
#     """
#     Test the conversion of a SAD SOL element with ref shifts at entry to XSuite.
#     """
#     reference_sol_test(
#         test_name                  = "test_007_weak_sol",
#         sad_elements_line_string   = textwrap.dedent(f"""\
#             DRIFT       SHORT_DRIFT = (L = 0.10);

#             SOL         SOL_IN      = (BZ = 0.001 BOUND =1  DX = TEST_VAL DY = TEST_VAL CHI1 = TEST_VAL CHI2 = TEST_VAL GEO = 1)
#                         SOL_OUT     = (BZ = 0.001 BOUND =1);

#             MARK        START       = ()
#                         END         = ()
#                         SOL_START   = ()
#                         SOL_END     = ();

#             LINE        SOL_DRIFT   = (SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT
#                 SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT
#                 SHORT_DRIFT SHORT_DRIFT)
#                         TEST_LINE   = (START
#                 SOL_IN SOL_START SOL_DRIFT SOL_END SOL_OUT END);
#             """),
#         sad_eval_marker            = "END",
#         test_values                = generate_symlog_array(-6, -1, 11),
#         static_val                 = STATIC_OFFSET,
#         plot                       = True)

################################################################################
# Sol (checking expansion of the square root)
################################################################################
def test_sol_on():
    """
    Test the conversion of a SAD SOL element with ref shifts at entry to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_007_sol",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SHORT_DRIFT = (L = 0.10);

            SOL         SOL_IN      = (BZ = 1.00 BOUND =1  DX = TEST_VAL DY = TEST_VAL DPX = TEST_VAL DPY = TEST_VAL GEO = 1)
                        SOL_OUT     = (BZ = 1.00 BOUND =1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        SOL_DRIFT   = (SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT
                SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT SHORT_DRIFT
                SHORT_DRIFT SHORT_DRIFT)
                        TEST_LINE   = (START
                SOL_IN SOL_START SOL_DRIFT SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = generate_symlog_array(-6, -1, 11),
        static_val                 = STATIC_OFFSET,
        plot                       = True)



"""
(Unofficial) SAD to XSuite Converter
"""

################################################################################
# Required Packages
################################################################################
import os
import sad2xs as s2x
import xtrack as xt
import numpy as np
import textwrap
import matplotlib.pyplot as plt
from tqdm import tqdm

from _sad_helpers import sad_twiss
from _test_config import *

################################################################################
# PyTest Function
################################################################################
def test_drift():
    """
    Test the conversion of a SAD DRIFT element to XSuite.
    """

    ############################################################################
    # Scan
    ############################################################################
    S_SAD       = np.zeros_like(POSITIVE_TEST_VALUES)
    S_XS        = np.zeros_like(POSITIVE_TEST_VALUES)
    X_SAD       = np.zeros_like(POSITIVE_TEST_VALUES)
    X_XS        = np.zeros_like(POSITIVE_TEST_VALUES)
    Y_SAD       = np.zeros_like(POSITIVE_TEST_VALUES)
    Y_XS        = np.zeros_like(POSITIVE_TEST_VALUES)
    PX_SAD      = np.zeros_like(POSITIVE_TEST_VALUES)
    PX_XS       = np.zeros_like(POSITIVE_TEST_VALUES)
    PY_SAD      = np.zeros_like(POSITIVE_TEST_VALUES)
    PY_XS       = np.zeros_like(POSITIVE_TEST_VALUES)

    for iteration, TEST_VAL in enumerate(tqdm(POSITIVE_TEST_VALUES)):

        ########################################################################
        # Write Test Lattice
        ########################################################################
        with open("test_lattice.sad", "w") as f:
            f.write(textwrap.dedent(f"""\
            MOMENTUM    = 1.0 GEV;
            TEST_VAL    = {TEST_VAL};

            DRIFT       TEST_DRIFT  = (L = TEST_VAL);

            MARK        START       = ()
                        END         = ();

            LINE        TEST_LINE   = (START TEST_DRIFT END);
            """))

        ########################################################################
        # Twiss SAD Lattice
        ########################################################################
        tw_sad  = sad_twiss(
            lattice_filename    = 'test_lattice.sad',
            line_name           = 'TEST_LINE')

        ########################################################################
        # Convert Lattice
        ########################################################################
        line    = s2x.convert_sad_to_xsuite(
            sad_lattice_path    = 'test_lattice.sad',
            output_directory    = None,
            _verbose            = False,
            _test_mode          = True)

        ########################################
        # Twiss XSuite Lattice
        ########################################
        tw_xs   = line.twiss4d(
            _continue_if_lost   = True,
            start               = xt.START,
            end                 = xt.END,
            betx                = 1,
            bety                = 1)

        ########################################################################
        # Save data for comparison
        ########################################################################
        S_SAD[iteration]    = tw_sad["s", "END"]
        S_XS[iteration]     = tw_xs["s", "end"]
        X_SAD[iteration]    = tw_sad["x", "END"]
        X_XS[iteration]     = tw_xs["x", "end"]
        Y_SAD[iteration]    = tw_sad["y", "END"]
        Y_XS[iteration]     = tw_xs["y", "end"]
        PX_SAD[iteration]   = tw_sad["px", "END"]
        PX_XS[iteration]    = tw_xs["px", "end"]
        PY_SAD[iteration]   = tw_sad["py", "END"]
        PY_XS[iteration]    = tw_xs["py", "end"]

        ########################################################################
        # Delete test lattice
        ########################################################################
        os.remove("test_lattice.sad")

    ############################################################################
    # Assertions
    ############################################################################
    assert np.all(
        np.isclose(S_SAD, S_XS, rtol = DELTA_S_RTOL, atol = DELTA_S_ATOL)), \
        "s values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(X_SAD, X_XS, rtol = DELTA_X_RTOL, atol = DELTA_X_ATOL)), \
        "x values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(Y_SAD, Y_XS, rtol = DELTA_Y_RTOL, atol = DELTA_Y_ATOL)), \
        "y values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(PX_SAD, PX_XS, rtol = DELTA_PX_RTOL, atol = DELTA_PX_ATOL)), \
        "px values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(PY_SAD, PY_XS, rtol = DELTA_PY_RTOL, atol = DELTA_PY_ATOL)), \
        "py values do not match between SAD and XSuite."

    ############################################################################
    # Plot
    ############################################################################
    symlog_threshold    = np.nanmin(
        np.where(POSITIVE_TEST_VALUES > 0, POSITIVE_TEST_VALUES, np.nan))

    if not np.all(np.isclose(S_SAD, S_XS, rtol = DELTA_S_RTOL, atol = DELTA_S_ATOL)):
        fig, axs    = plt.subplots(
            nrows       = 2,
            ncols       = 1,
            figsize     = (12, 6),
        sharex      = True,
        gridspec_kw = {'height_ratios': [3, 1]})

        axs[0].plot(POSITIVE_TEST_VALUES, S_SAD, label = "SAD", color = "r")
        axs[0].plot(POSITIVE_TEST_VALUES, S_XS, label = "XS", color = "b", linestyle='--')
        axs[1].plot(POSITIVE_TEST_VALUES, S_XS - S_SAD)

        axs[1].set_xlabel("Drift Length [m]")
        axs[0].set_ylabel('s [m]')
        axs[1].set_ylabel('∆s [m]')

        for ax in axs:
            ax.set_xscale('symlog', linthresh = symlog_threshold)
            ax.set_yscale('symlog', linthresh = symlog_threshold)
            ax.grid()
        axs[0].legend()
        plt.subplots_adjust(hspace = 0)
        plt.savefig(f"outputs/test_001_drift_s.png", dpi = 300, bbox_inches = 'tight')

    plt.close("all")
